#!/bin/python
from typing import Dict, Tuple
from shap import Explanation
from argueview.typings import FeatureImportance, Case
import numpy as np


def feature_importance_from_lime(explanation: Dict[int, FeatureImportance], case: Case) -> Tuple[
    FeatureImportance, float]:
    # invert feature contributions when they are negatively framed w.r.t. decision-class
    for i in range(0, len(explanation[1])):
        v = list(explanation[1][i])
        v[1] = -v[1]
        explanation[1][i] = tuple(v)

    # extract feature contribution map
    feature_importance = explanation[1]

    # compute unexplained
    change = sum(list(map(lambda x: x[1], explanation[1])))
    final = max(case.class_proba)
    unexplained = final - change

    return feature_importance, unexplained


def feature_importance_from_shap(shap_values: Explanation, case: Case) -> Tuple[FeatureImportance, float]:
    decision_class = np.argmax(case.class_proba)
    shap_values_for_class = shap_values[..., decision_class]

    # save as feature_importance map
    dat = iter(shap_values_for_class[case.id].values)
    fid = iter([i for i in range(len(shap_values_for_class[case.id].values))])
    feature_importance = list(zip(fid, dat))

    # unexplained importance
    unexplained = shap_values_for_class[case.id].base_values

    return feature_importance, unexplained



import os
import io
try:
    from IPython.core.display import display, HTML, Javascript
    from IPython import get_ipython
    have_ipython = True
except ImportError:
    have_ipython = False


def html_err_msg() -> str:
    return """
    <div style='color: #900; text-align: left;'>
      <b>Visualization omitted, Javascript library not loaded!</b><br>
      If this notebook was from another user you must trust this notebook (File -> Trust notebook). 
      If you are viewing this notebook on github the Javascript has been stripped for security. If you are using
      JupyterLab this error is because a JupyterLab extension has not yet been written.
    </div>"""


def _ipython_err_msg() -> str:
    return "IPython must be installed to use initjs()! Run `pip install ipython` and then restart argueview."


def initjs() -> None:
    assert have_ipython, _ipython_err_msg()
    bundle_path = os.path.join(os.path.split(__file__)[0], "resources", "bundle.min.js")
    with io.open(bundle_path, encoding="utf-8") as f:
        bundle_data = f.read()
    display(Javascript("{bundle_data}".format(bundle_data=bundle_data)))
    while True:
        yield



import textwrap
from random import random
from typing import List, Union, NewType, Tuple
import jsonpickle


class DummyUpdater(object):
    def __init__(self, iterable=(), **kwargs):
        self.__dict__.update(iterable, **kwargs)


class Author(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    name: str
    organization: str


class Feature(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    data_type: str
    description: str
    index: int
    is_ignore: bool
    is_row_identifier: bool
    is_target: bool
    name: str
    nominal_value: List[str]
    number_of_missing_values: int

    def __setitem__(self, key:str, value:any) -> None:
        self[key] = value


class Source(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    name: str
    author: Author
    description: str
    href: str
    observations: int
    type: str
    year: int
    features: List[Feature]


class OpenMLFeatureData(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    feature: List[Feature]


class CaseFeature(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    value: Union[float, str]


class CaseSource(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    features: List[CaseFeature]


class Case(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    id: int
    class_proba: List[float]
    sources: List[CaseSource]

    def decision_class(self) -> int:
        return self.class_proba.index(max(self.class_proba))

    def qualifier(self) -> float:
        proba_decision = self.class_proba[self.decision_class()]
        proba_uncertain = 1 / len(self.class_proba)
        full_range = (1 - proba_uncertain)
        pos = proba_decision - proba_uncertain
        return pos / full_range


class LatentContinuousTarget(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    label: str
    anti_label: str
    mapping: List[float]


class Data(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    classes: List[str]
    latent_continuous_target: LatentContinuousTarget
    sources: List[Source]


class FeatureExplanation(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    source: int
    feature: int
    value: str
    contribution: float


class ExplanationPartial(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    base: float
    support: List[FeatureExplanation]
    attack: List[FeatureExplanation]

    def find(self, source: int, feature: int) -> Union[FeatureExplanation, None]:
        l: List[FeatureExplanation] = [*self.support, *self.attack]
        for i in range(0, len(l)):
            if l[i].source == source and l[i].feature == feature:
                return l[i]
        return None


class Color:
    BOLD = '\033[1m'
    END = '\033[0m'


class Explanation(DummyUpdater):
    def __init__(self, iterable=(), **kwargs):
        super().__init__(iterable, **kwargs)

    backing: str
    case: Case
    data: Data
    explanation: ExplanationPartial

    def serialize(self) -> str:
        return jsonpickle.encode(self, unpicklable=False)

    def save(self, fname: str) -> None:
        json_out = open(fname, 'w')
        json_out.write(self.serialize())
        json_out.close()

    def feature_value(self, source: int, feature: int) -> Union[str, int, float]:
        if source >= len(self.case.sources) or feature >= len(self.case.sources[source].features):
            return 0
        val = self.case.sources[source].features[feature].value
        fdata = self.data.sources[source].features[feature]
        if fdata.data_type == 'nominal':
            return fdata.nominal_value[int(val)]
        else:
            return val

    def _print_multiline(self, inpstr: str, str_pre: str, str_post: str, str_max_length: int, max_length: int, cl: int) -> None:
        if len(inpstr) - cl <= max_length:
            print(str_pre + inpstr + ' ' * (str_max_length - len(inpstr) + cl) + str_post)
        else:
            inpstr_lines = textwrap.wrap(inpstr, max_length)
            for i in range(0, len(inpstr_lines)):
                print(str_pre + inpstr_lines[i] + ' ' * (
                            str_max_length - len(inpstr_lines[i]) + (cl if i == 0 else 0)) + str_post)

    def print(self, plain=False) -> None:

        # compute variables
        dclass = self.data.classes[self.case.decision_class()]
        support_max = 0
        support_i = 0
        for i in range(0, len(self.explanation.support)):
            if self.explanation.support[i].contribution > support_max:
                support_max = self.explanation.support[i].contribution
                support_i = i
        lrat = self.explanation.support[support_i].value

        if plain:
            print('Explanation')
            print('class: '+dclass)
            print('leading rationale:'+lrat)
            print('qualifier: '+"The class '" + dclass + "' is {:.0f}% more certain than the other possible classes.".format(100 * self.case.qualifier()))
            print('backing: '+self.backing)
            print()
            for i in range(0, len(self.case.sources)):
                print('Source: '+self.data.sources[i].name)
                names = []
                vals = []
                cont = []
                contval = []
                for j in range(0, len(self.data.sources[i].features)):
                    ef = self.explanation.find(i, j)
                    names.append(self.data.sources[i].features[j].name)
                    vals.append(str(self.feature_value(i, j)))
                    cont.append((str(ef.contribution) if ef is not None else "None"))
                    contval.append(ef.contribution if ef is not None else -999999)
                names_max_length = max(map(lambda x: len(x), names))
                vals_max_length = max(map(lambda x: len(x), vals))
                for j in range(0, len(names)):
                    print(names[j] + ' ' * (names_max_length - len(names[j])) + ' = ' + vals[j] + ' ' * (
                            vals_max_length - len(vals[j])) + ' , contribution = ' + cont[j])
                print()
            print('Unexplained contribution: ' + str(self.explanation.base))

        else:
            # compute strings
            str_header = '-- Explanation '
            str_pre = '|  '
            str_post = '  |'
            str_class = Color.BOLD + 'class: ' + Color.END + dclass
            str_lrat = Color.BOLD + 'leading rationale: ' + Color.END + lrat
            str_qualifier = Color.BOLD + "qualifier: " + Color.END + "The class '" + dclass + "' is {:.0f}% more certain than the other possible classes.".format(
                100 * self.case.qualifier())
            str_backing = Color.BOLD + 'backing: ' + Color.END + self.backing

            # print explanation
            max_length = 80
            cl = len(Color.BOLD) + len(Color.END)
            str_max_length = min(max_length, max(len(str_class), len(str_lrat), len(str_qualifier), len(str_backing)))
            print('\n' + str_header + '-' * (str_max_length - len(str_header) + len(str_pre) + len(str_post)))
            self._print_multiline(str_class, str_pre, str_post, str_max_length, max_length, cl)
            self._print_multiline(str_lrat, str_pre, str_post, str_max_length, max_length, cl)
            self._print_multiline(str_qualifier, str_pre, str_post, str_max_length, max_length, cl)
            self._print_multiline(str_backing, str_pre, str_post, str_max_length, max_length, cl)
            print('-' * (str_max_length + len(str_pre) + len(str_post)) + '\n')

            for i in range(0, len(self.case.sources)):

                # compute strings for source
                str_header = '-- ' + self.data.sources[i].name + ' '
                names = []
                vals = []
                cont = []
                contval = []
                for j in range(0, len(self.data.sources[i].features)):
                    ef = self.explanation.find(i, j)
                    names.append(self.data.sources[i].features[j].name)
                    vals.append(str(self.feature_value(i, j)))
                    cont.append((str(ef.contribution) if ef is not None else "None"))
                    contval.append(ef.contribution if ef is not None else -999999)
                names_max_length = max(map(lambda x: len(x), names))
                vals_max_length = max(map(lambda x: len(x), vals))
                content = []
                for j in range(0, len(names)):
                    content.append(names[j] + ' ' * (names_max_length - len(names[j])) + ' = ' + vals[j] + ' ' * (
                                vals_max_length - len(vals[j])) + ' , contribution = ' + cont[j])
                str_max_length = max(map(lambda x: len(x), content))

                # sort based on contribution
                content = list(zip(*sorted(zip(contval, content), reverse=True)))[1]

                # print source
                print(str_header + '-' * (str_max_length - len(str_header) + len(str_pre) + len(str_post)))
                for j in range(0, len(content)):
                    print(str_pre + content[j] + ' ' * (str_max_length - len(content[j])) + str_post)
                print('-' * (str_max_length + len(str_pre) + len(str_post)))
            print('Unexplained contribution: ' + str(self.explanation.base))


Ground = NewType('Ground', List[str])
Grounds = NewType('Grounds', List[Ground])
GroundVarByClass = NewType('GroundVarByClass', List[str])
GroundsVarByClass = NewType('GroundsVarByClass', List[GroundVarByClass])
FeatureImportance = NewType('FeatureImportance', List[Tuple[int, float]])



from typing import List, Union
from argueview.typings import Source, Explanation, Case, Data, ExplanationPartial, FeatureExplanation, Grounds, \
    GroundsVarByClass, FeatureImportance, Feature, LatentContinuousTarget


class ArgueView:
    """Argumentative explanation presentation tool for machine-learning and decision-support systems.

    ArgueView provides functionality to organize contextual information about your data sources, the decision-classes,
    feature-importance based explanations, and a decision backing. With provided information ArgueView can generate a
    json-styled data explanation object that can be exported and used in visualization tools.
    """

    _backing: str
    _sources: List[Source]
    _classes: List[str]
    _lct: LatentContinuousTarget
    _grounds: Grounds
    _grounds_vars: GroundsVarByClass

    _case: Case
    _fimportance: List[FeatureImportance]
    _unexplained: float
    _gen: Explanation

    _expPartial: ExplanationPartial

    def __init__(self) -> None:
        self._sources = []

    def classes(self, classes: List[str]) -> None:
        """Defines textual representations for the decision-classes.

        For example, your decision task is a binary classification of pears and oranges, encoded as 0 or 1.
        You should then call this method with human-interpretable text representations for each class.
        An example would be classes = ['pears', 'orange']

        :param  classes: Array of textual representations for your decision-classes.
        :return:
        """
        if self._lct and len(self._lct.mapping) != len(classes):
            raise Exception('The number of classes should match the length of the defined latent continuous target mapping.')
        self._classes = classes

    def latent_continuous_target(self, label: str, anti_label: str, mapping: List[float]) -> None:
        """Define a latent continuous target present within your decision-classes.

        To illustrate the usage of the latent continuous target assume a decision task in the loan application domain
        where predictions determine whether someone is eligible for a loan. The decision-classes are: 'applicable for loan'
        and 'not applicable for loan'. In this scenario the latent continuous target is 'applicability' and it's anti-label
        is 'inapplicability'. The mapping is [1, -1] which indicates the value of the latent target for each decision-class.

        :param label: Label for the latent continuous target.
        :param anti_label: Anti-label for the latent continuous target.
        :param mapping: List of values representing the value of the latent target for each decision-class.
        :return:
        """
        if self._classes and len(self._classes) != len(mapping):
            raise Exception('Length of mapping should match the number of classes.')
        self._lct = LatentContinuousTarget({
            label: label,
            anti_label: anti_label,
            mapping: mapping
        })

    def backing(self, backing: str) -> None:
        """Defines textual backing to support your decisions.

        A strong backing should describe the data quality and the quality of the provider.

        :param  backing: A textual backing to support your decisions.
        :return:
        """

        self._backing = backing

    def grounds(self, grounds: Grounds, grounds_var_by_class: GroundsVarByClass) -> None:
        """Define grounds for the rationale of how each feature affects the decision-class.

        The grounds describe how features affect the decision in each possible decision-class. The grounds can contain
        tokens that should be replaced based on the feature value. The token used for this is `<>`.

        :param  grounds: Contains for each feature the rationale for each decision class. The format is List[List[str]] where the outer list represents each feature and the inner list each decision-class.
        :param  grounds_var_by_class:
        :return:
        """

        self._grounds = grounds
        self._grounds_vars = grounds_var_by_class

    def add_data_source(self, source: Source) -> None:
        self._sources.append(source)
        if not isinstance(source.features[0], Feature):
            for i in range(0, len(source.features)):
                self._sources[len(self._sources)-1].features[i] = Feature(self._sources[len(self._sources)-1].features[i])

    def _compile(self, decision: int, source: int, feature: int) -> str:
        """Compiles a rationale for a decision from a specific feature.

        :param  decision: The decision-class.
        :param  source: Data source.
        :param  feature: Feature number.
        :return: A compiled rationale for the contribution of `feature` to the decision-class. Based on the grounds provided in ArgueView.grounds()
        """

        ground = self._grounds[feature][decision]
        feature_value = int(self._case.sources[source].features[feature].value)
        options = self._grounds_vars[feature]
        if feature_value >= len(options):
            return ground
        return ground.replace("<>", options[feature_value])

    def _generate_explanation_partial(self) -> None:
        positive = []
        negative = []
        decision = self._case.decision_class()
        for i in range(0, len(self._sources)):
            for t in self._fimportance[i]:
                valence = negative
                if t[1] > 0:
                    valence = positive
                valence.append(FeatureExplanation({
                    "source": i,
                    "feature": int(t[0]),
                    "contribution": float(t[1]),
                    "value": self._compile(decision, i, t[0])
                }))
        self._expPartial = ExplanationPartial({
            "support": positive,
            "attack": negative,
            "base": float(self._unexplained)
        })

    def generate(self, case: Case, feature_importance: Union[FeatureImportance, List[FeatureImportance]], unexplained: float = 0) -> Explanation:
        """Generate an explanation for a case based on the given context.

        Generates an explanation for a specific case. This method uses the information provided by other methods and
        assembles this information and the current case into an explanation object. This object can be used to
        visualize the explanation or to export it for usage in other tools.

        :param case: Case object containing the class probabilities and feature values.
        :param feature_importance: Feature importance for the case as computed by an explainer. A list of FeatureImportance
        objects is required for multi-source scenarios.
        :param unexplained: Unexplained contribution for the decision-class.
        :return:
        """

        pre = 'cannot generate explanation: '
        if not self._classes:
            raise Exception(pre+'no classes defined. Use ArgueView.classes to set your classes.')
        if not self._grounds:
            raise Exception(pre+'no grounds defined. Use ArgueView.grounds to set your grounds.')
        if not self._backing:
            print('Warning: Generating explanation without a backing.')
        if not self._sources or len(self._sources) <= 0:
            raise Exception(pre+'no data sources defined. Use ArgueView.add_data_source to add a data source.')

        if not isinstance(feature_importance[0], list):
            self._fimportance = [feature_importance] # single list -> single source case
        else:
            self._fimportance = feature_importance # input properly formatted
        self._unexplained = unexplained

        self._case = case
        self._generate_explanation_partial()
        _gen = Explanation()

        if self._backing:
            _gen.backing = self._backing

        _gen.data = Data()
        _gen.data.classes = self._classes
        _gen.data.sources = self._sources
        _gen.data.latent_continuous_target = self._lct
        _gen.case = self._case
        _gen.explanation = self._expPartial
        return _gen



import setuptools
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    install_requires=[
        "backcall==0.2.0",
        "decorator==4.4.2",
        "importlib-metadata==2.0.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
        "ipython==7.19.0",
        "ipython-genutils==0.2.0",
        "jedi==0.17.2; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
        "jsonpickle==1.4.1",
        "numpy==1.19.4",
        "parso==0.7.1; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "pexpect==4.8.0; sys_platform != 'win32'",
        "pickleshare==0.7.5",
        "prompt-toolkit==3.0.8; python_full_version >= '3.6.1'",
        "ptyprocess==0.6.0",
        "pygments==2.7.2; python_version >= '3.5'",
        "scipy==1.5.4",
        "traitlets==5.0.5; python_version >= '3.7'",
        "wcwidth==0.2.5",
        "zipp==3.4.0; python_version >= '3.6'",
    ],
    name="argueview",
    version="0.1.4",
    author="Sophia Hadash",
    author_email="s.hadash@tue.nl",
    description="ArgueView is a tool for generating text-based presentations for machine-learning predictions and feature-importance based explanation tools. The tool makes use of Toulmin's model of argumentation for structuring the text-based explanations.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SophiaHadash/ArgueView",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Development Status :: 3 - Alpha",
    ],
    keywords="explanations, text, toulmin, argumentation",
    python_requires=">=3.5",
)



from random import random
from IPython.core.display import display, Javascript, HTML
from argueview.typings import Explanation
from argueview.visualizers import initjs, html_err_msg


def ToulminVisualizer(explanation: Explanation) -> None:
    next(initjs())
    eid = "vis-" + str(random())
    display(HTML("<div id='{id}'>{err_msg}</div>".format(id=eid, err_msg=html_err_msg())))
    display(Javascript("""
    if (window.argueview) {{
        window.ReactDom.render(
            window.React.createElement(window.argueview.ToulminVisualizer, {{explanation: {data}}}),
            document.getElementById('{id}')
        );
    }}
    """.format(id=eid, data=explanation.serialize())))
